# -*- coding:utf-8 -*-
# CREATED BY: jiangbohuai
# CREATED ON: 2021/11/26 4:48 PM
# LAST MODIFIED ON:
# AIM:

from typing import List


def flatten(array: List) -> List:
    out = []
    for v in array:
        if type(v) == list:
            out.extend(flatten(v))
        else:
            out.append(v)
    return out



# -*- coding:utf-8 -*-
# CREATED BY: jiangbohuai
# CREATED ON: 2021/11/11 11:02 AM
# LAST MODIFIED ON:
# AIM:
import re
from typing import List

from sentence_spliter.architect.sequence import SequenceABC
from sentence_spliter.en_cutter.symbol import Symbols
from sentence_spliter.utility.arrary import flatten


class Sequence(SequenceABC):
    def __init__(self, str_block: str):
        super(Sequence, self).__init__(str_block)

        self.n_words = 0
        # self.word_pattern = re.compile(r'[a-zA-Z0-9]+')

    def next(self):
        # if self.word_pattern.match(self.current_token):
        #     self.n_words += 1
        self.n_words += 1
        super(Sequence, self).next()

    def back(self):
        # if self.word_pattern.match(self.current_token):
        #     self.n_words = max(0, self.n_words - 1)
        self.n_words = max(0, self.n_words - 1)
        super(Sequence, self).back()

    def add_to_results(self):
        self.n_words = -1
        super(Sequence, self).add_to_results()

    def tokenizer(self, str_block: str) -> List[str]:
        return list(str_block)

    def update_quota(self):

        if self.current_token in Symbols.s_quota_en.value:
            self.s_quota_en = (self.s_quota_en + 1) % 2
        if self.current_token in Symbols.s_quota_left.value:
            self.s_quota_left += 1
        if self.current_token in Symbols.s_quota_right.value:
            self.s_quota_right += -1
        if self.current_token in Symbols.quotation_en.value:
            self.quota_en = (self.quota_en + 1) % 2
        if self.current_token in Symbols.quotation_left.value:
            self.quota_left += 1
        if self.current_token in Symbols.quotation_right.value:
            self.quota_right += -1

    def degrade_quota(self):
        if self.current_token in Symbols.all_quota.value or \
                self.current_token in Symbols.all_s_quota.value:
            if self.current_token in Symbols.s_quota_en.value:
                self.s_quota_en = (self.s_quota_en - 1) % 2
            if self.current_token in Symbols.s_quota_left.value:
                self.s_quota_left -= 1
            if self.current_token in Symbols.s_quota_right.value:
                self.s_quota_right += 1
            if self.current_token in Symbols.quotation_en.value:
                self.quota_en = (self.quota_en - 1) % 2
            if self.current_token in Symbols.quotation_left.value:
                self.quota_left -= 1
            if self.current_token in Symbols.quotation_right.value:
                self.quota_right += 1

    def update_bracket(self):
        if self.current_token in Symbols.bracket_right.value:
            self.bracket_right -= 1
        if self.current_token in Symbols.bracket_left.value:
            self.bracket_left += 1

    def degrade_bracket(self):
        if self.current_token in Symbols.bracket_right.value:
            self.bracket_right += 1
        if self.current_token in Symbols.bracket_left.value:
            self.bracket_left -= 1



# -*- coding:utf-8 -*-
# CREATED BY: jiangbohuai
# CREATED ON: 2021/11/15 3:21 PM
# LAST MODIFIED ON:
# AIM:
from typing import List, Union
import re
import pathlib
import os

from sentence_spliter.architect.graph_component import Condition
from sentence_spliter.en_cutter.symbol import Symbols
from sentence_spliter.en_cutter.en_sequence import Sequence
from sentence_spliter.utility.file_opt import read_file
from sentence_spliter.utility.trie import Trie

PATH = str(pathlib.Path(__file__).absolute().parent)
WHITE_LIST_PATH = os.path.join(PATH, 'white_list.txt')


class IsEndState(Condition):
    def __init__(self) -> object:
        super(IsEndState, self).__init__('IsEndState')

    def run(self, tok: Sequence):
        return tok.current_token == ''


class IsEndSymbol(Condition):
    def __init__(self, name: str = 'IsEndSymbol', reverse: bool = False):
        super(IsEndSymbol, self).__init__(name, reverse)

    def run(self, tok: Sequence) -> bool:
        return tok.current_token in Symbols.end_symbols.value


class IsBracketClose(Condition):
    def __init__(self, name: str = 'IsBracketClose', reverse: bool = False):
        super(IsBracketClose, self).__init__(name, reverse)

    def run(self, tok: Sequence) -> bool:
        if tok.bracket_left + tok.bracket_right == 0:
            tok.reset_bracket()
            return True
        return False


class IsQuotaClose(Condition):
    def __init__(self, name: str = 'IsQuotaClose', reverse: bool = False):
        super(IsQuotaClose, self).__init__(name, reverse)

    def run(self, tok: Sequence) -> bool:
        if tok.current_token in Symbols.quotation_right.value:
            return True
        if abs(tok.quota_right + tok.quota_left) + tok.quota_en == 0:
            return True
        return False


class IsSingleQuotaClose(Condition):
    def __init__(self, name: str = 'IsSingleQuotaClose', reverse: bool = False):
        super(IsSingleQuotaClose, self).__init__(name, reverse)

    def run(self, tok: Sequence) -> bool:
        if abs(tok.s_quota_left + tok.s_quota_right) + tok.s_quota_en == 0:
            tok.reset_s_quota()
            return True
        return False


class IsBlank(Condition):
    def __init__(self, name: str = 'IsBlank', reverse: bool = False):
        super(IsBlank, self).__init__(name, reverse)
        self.pattern = re.compile(r'^\s+$')

    def run(self, tok: Sequence) -> bool:
        if self.pattern.match(tok.current_token):
            return True
        else:
            return False


class TokenInWhiteList(Condition):
    def __init__(self, name: str = 'TokenInWhiteList', reverse: bool = False):
        super(TokenInWhiteList, self).__init__(name, reverse)
        self.white_list = Trie('<root>')
        for value in read_file(WHITE_LIST_PATH).splitlines():
            if not re.match('[a-zA-Z]', value[-1]):
                token_list = value[0:-1].split() + [value[-1]]
            else:
                token_list = value.split()
            self.white_list.add_phrase(token_list[::-1])

    def run(self, tok: Sequence) -> bool:
        verb = tok.current_token
        i = tok.i - 1
        tree = self.white_list
        while verb in tree:
            if re.match(r'^\s+$', verb):
                continue
            tree = tree[verb]
            verb = tok[i]
            i -= 1

        if tree.is_end_node(tok[i]):
            return True
        return False


class IsRightQuota(Condition):
    def __init__(self, name: str = 'IsRightQuota', reverse: bool = False):
        super(IsRightQuota, self).__init__(name, reverse)

    def run(self, seq: Sequence) -> bool:
        tok = seq.current_token
        if tok in Symbols.quotation_right.value:
            return True
        if seq.quota_en == 0 and seq.current_token in Symbols.quotation_en.value:
            return True
        return False


class IsLeftQuota(Condition):
    def __init__(self, name: str = 'IsLeftQuota', reverse: bool = False):
        super(IsLeftQuota, self).__init__(name, reverse)

    def run(self, seq: Sequence) -> bool:
        tok = seq.current_token
        if tok in Symbols.quotation_left.value:
            return True
        if seq.quota_en == 1 and seq.current_token in Symbols.quotation_en.value:
            return True
        return False


class IsRightBeforeLeftQuota(Condition):
    def __init__(self, name: str = 'IsRightBeforeLeftQuota', reverse: bool = False):
        super(IsRightBeforeLeftQuota, self).__init__(name, reverse)
        self.right_quota = IsRightQuota()

    def run(self, seq: Sequence):
        next_tok = seq.get_right_nearest_tok()
        if self.right_quota(seq) and (
                next_tok in Symbols.quotation_left.value or next_tok in Symbols.quotation_en.value):
            return True
        return False


class IsRightSingleQuota(Condition):
    def __init__(self, name: str = "IsRightSingleQuota", reverse: bool = False):
        super(IsRightSingleQuota, self).__init__(name, reverse)

    def run(self, seq: Sequence) -> bool:
        tok = seq.current_token
        if tok in Symbols.s_quota_right.value:
            return True
        if seq.quota_en == 0 and seq.current_token in Symbols.s_quota_en.value:
            return True
        return False


class IsLeftSingleQuota(Condition):
    def __init__(self, name: str = "IsLeftSingleQuota", reverse: bool = False):
        super(IsLeftSingleQuota, self).__init__(name, reverse)

    def run(self, seq: Sequence) -> bool:
        tok = seq.current_token
        if tok in Symbols.s_quota_left.value:
            return True
        if seq.quota_en == 1 and seq.current_token in Symbols.s_quota_en.value:
            return True
        return False


class IsNumberDot(Condition):
    """
    case 1. go to bed.
    or   1.1. go to bed.

    1. may not be a sentence 
    """

    def __init__(self, name: str = 'IsNumberDot', reverse: bool = False):
        super(IsNumberDot, self).__init__(name, reverse)
        self.pattern = re.compile('^\d+(.\d+)*$')

    def run(self, tok: Sequence) -> bool:
        if tok.current_token == '.' and self.pattern.match(tok[tok.i-1]):
            return True
        return False


class NextStartWithCapital(Condition):
    def __init__(self, name: str = 'NextStartWithCapital', reverse: bool = False):
        super(NextStartWithCapital, self).__init__(name, reverse)
        self.pattern = re.compile(r'^[A-Za-z]+')

    def run(self, tok: Sequence) -> bool:
        return bool(self.pattern.match(tok[tok.i + 2]))


class NextStartWithNum(Condition):
    def __init__(self, name: str = 'NextStartWithNum', reverse: bool = False):
        super(NextStartWithNum, self).__init__(name, reverse)
        self.pattern = re.compile(r'^\d')

    def run(self, tok: Sequence) -> bool:
        return bool(self.pattern.match(tok[tok.i + 2]))


class IsSentenceDash(Condition):
    def __init__(self, name: str = 'IsSentenceDash', reverse: bool = False):
        super(IsSentenceDash, self).__init__(name, reverse)

    def run(self, tok: Sequence) -> bool:
        if tok.current_token in Symbols.dash.value:
            return True
        if tok.current_token in Symbols.short_dash.value and tok[tok.i - 1] in Symbols.short_dash.value:
            return True
        return False


class IsComma(Condition):
    def __init__(self, name: str = 'IsComma', reverse: bool = False):
        super(IsComma, self).__init__(name, reverse)

    def run(self, tok: Sequence) -> bool:
        if tok.current_token in Symbols.comma.value:
            return True
        return False


# --- long short handler -- #
class IsLongSentence(Condition):
    def __init__(self, max_length: int = 40, name: str = 'IsLongSentence', reverse: bool = False):
        super(IsLongSentence, self).__init__(name, reverse)
        self.max_length = max_length

    def run(self, tok: Sequence) -> bool:
        return tok.n_words > self.max_length


class IsShortSentence(Condition):
    def __init__(self, min_length: int = 6, name: str = 'IsShortSentence', reverse: bool = False):
        super(IsShortSentence, self).__init__(name, reverse)
        self.min_length = min_length

    def run(self, tok: Sequence) -> bool:
        return tok.n_words < self.min_length


class RightQuotaAfterEnd(Condition):
    def __init__(self, name='RightQuotaAfterEnd', reverse: bool = False):
        super(RightQuotaAfterEnd, self).__init__(name, reverse)
        self.is_right_quota = IsRightQuota().add_or(IsRightSingleQuota())

    def run(self, tok: Sequence) -> bool:
        if self.is_right_quota(tok):
            if tok[tok.i - 1] in Symbols.end_symbols.value:
                return True
        return False


class IsNearestLeftQuota(Condition):
    def __init__(self, name='IsNearestRightQuota', reverse: bool = False):
        """
        case  'he said. "... ' > ['he said. ','"...']
        :param name:
        :param reverse:
        """
        super(IsNearestLeftQuota, self).__init__(name, reverse)

    def run(self, seq: Sequence) -> bool:
        tok = seq.get_right_nearest_tok()
        if tok in Symbols.quotation_left.value or tok in Symbols.s_quota_left.value:
            return True
        if (seq.quota_en == 0 and tok in Symbols.quotation_en.value) \
                or (seq.s_quota_en == 0 and tok in Symbols.s_quota_en.value):
            return True
        return False



# -*- coding:utf-8 -*-
# CREATED BY: jiangbohuai
# CREATED ON: 2021/11/16 9:56 AM
# LAST MODIFIED ON:
# AIM:

def read_file(file_path: str) -> str:
    with open(file_path, 'r') as f:
        return f.read()



# -*- coding:utf-8 -*-
# CREATED BY: jiangbohuai
# CREATED ON: 2021/10/12 5:40 PM
# LAST MODIFIED ON:
# AIM:

from typing import List
import abc

import re


class SequenceABC:
    def __init__(self, str_block: str):
        self.tokens = self.tokenizer(str_block)
        # -- pointers -- #
        self.__i = 0
        self.sentence_start = 0
        # --- bracket --- #
        self.bracket_left = 0
        self.bracket_right = 0
        # --- quota --- #
        self.quota_left = 0
        self.quota_right = 0
        self.quota_en = 0
        self.s_quota_en = 0
        self.s_quota_left = 0
        self.s_quota_right = 0

        # -- results -- #
        self.sentence_list_idx = []

        # -- init -- #
        if self.__i == 0:
            self.update_quota()
            self.update_bracket()

    def __str__(self):
        return self.current_token

    def __getitem__(self, item):
        if -len(self) <= item < len(self):
            return self.tokens[item]
        else:
            return ''

    def __len__(self):
        return len(self.tokens)

    def __iter__(self):
        while self.current_token:
            yield self
            self.next()

    @property
    def current_token(self):
        return self[self.__i]

    @property
    def i(self) -> int:
        return self.__i

    @i.setter
    def i(self, value: int):
        # forward
        if value > self.__i:
            while self.__i != min(value, self.__len__()):
                self.next()
        # backward
        else:
            while self.__i != max(0, value):
                self.back()

    def get_result(self):
        out = []
        for v in self.sentence_list_idx:
            sentence = ''.join(self.tokens[v[0]:v[1]])
            if sentence:
                out.append(sentence)
        return out

    def get_right(self) -> str:
        i = self.__i + 1
        tok = self[i]
        return tok

    def get_left(self) -> str:
        i = self.__i - 1
        tok = self[i]
        return tok

    def get_left_interval(self, length: int = 1) -> str:
        string = ''
        i = self.__i
        for _ in range(length):
            i -= 1
            string = self[i] + string
        return string

    def get_right_interval(self, length: int = 1) -> str:
        string = ''
        i = self.__i
        for _ in range(length):
            i += 1
            string = string + self[i]
        return string

    def get_right_nearest_tok(self) -> str:
        i = self.__i + 1
        tok = self[i]
        length = len(self)
        while re.match(r'^\s+$', tok) and i < length:
            i += 1
            tok = self[i]
        return tok

    def get_left_nearest_tok(self) -> str:
        i = self.__i - 1
        tok = self[i]
        while re.match(r'^\s+$', tok) and i > 0:
            i -= 1
            tok = self[i]
        return tok

    def reset_bracket(self):
        self.bracket_right = 0
        self.bracket_left = 0

    def reset_quota(self):
        self.quota_right = 0
        self.quota_left = 0

    def reset_s_quota(self):
        self.s_quota_left = 0
        self.s_quota_right = 0

    # --- abstract layer -- #
    @abc.abstractmethod
    def tokenizer(self, str_block: str) -> List[str]:
        pass

    @abc.abstractmethod
    def update_quota(self):
        pass

    @abc.abstractmethod
    def degrade_quota(self):
        pass

    @abc.abstractmethod
    def update_bracket(self):
        pass

    @abc.abstractmethod
    def degrade_bracket(self):
        pass

    # -- basic action -- #
    def next(self):
        self.__i += 1
        self.update_bracket()
        self.update_quota()

    def back(self):
        self.degrade_bracket()
        self.degrade_quota()
        self.__i -= 1

    def add_to_results(self):
        self.sentence_list_idx.append((self.sentence_start, self.__i + 1))
        self.sentence_start = self.__i + 1
        self.next()
